package cz.sa.shiftPlanner.daoImpl;

import java.util.Arrays;
import java.util.List;

import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.springframework.stereotype.Repository;

import cz.sa.shiftPlanner.dao.BusConnectionDAO;
import cz.sa.shiftPlanner.entities.BusConnectionVO;
import cz.sa.shiftPlanner.entities.LineGroupVO;
import cz.sa.shiftPlanner.entities.StationVO;
import cz.sa.shiftPlanner.enums.WeekDay;
import cz.sa.shiftPlanner.infrastructure.dao.AbstractDao;
import cz.sa.shiftPlanner.utils.DateFormatter;
import cz.sa.shiftPlanner.utils.TimeRange;

@Repository
public class BusConnectionDAOImpl extends AbstractDao implements BusConnectionDAO {

  private static final Logger log = Logger.getLogger(BusConnectionDAOImpl.class);

  @Override
  public BusConnectionVO get(Long id) {

    if (id == null)
      throw new IllegalArgumentException("id is null");
    Object o = getSession().get(BusConnectionVO.class, id);
    return (BusConnectionVO) o;
  }

  @Override
  public void create(BusConnectionVO connection) {
    if (connection == null)
      throw new IllegalArgumentException("connection");
    if (connection.getId() != null)
      throw new IllegalArgumentException("Already persisted");

    getSession().save(connection);

  }

  @Override
  public void update(BusConnectionVO connection) {
    if (connection == null)
      throw new IllegalArgumentException("connection");
    if (connection.getId() == null)
      throw new IllegalArgumentException("instance not persisted");

    getSession().update(connection);
    log.info("Connection " + connection.getId() + " was updated. Skupina linek "+connection.getTimetable().getLine().getName());
  }

  @Override
  public void remove(BusConnectionVO connection) {
    if (connection == null)
      throw new IllegalArgumentException("connection");
    if (connection.getId() == null)
      throw new IllegalArgumentException("instance not persisted");

    getSession().delete(connection);
  }

  @Override
  public List<BusConnectionVO> findAll() {
    return getSession().createQuery("FROM BusConnectionVO ").list();
  }

  @Override
  public List<BusConnectionVO> findConnectionByTime(DateTime departure, DateTime arrival) {
    return getSession()
        .createQuery("SELECT c FROM BusConnectionVO AS c WHERE c.departureTime >= :departure AND c.arrivalTime <= :arrival")
        .setParameter("departure", departure).setParameter("arrival", arrival).list();
  }

  @Override
  public BusConnectionVO findConnectionByConcreteFigure(DateTime departure, DateTime arrival, StationVO departureStation,
      StationVO arrivalStation, LineGroupVO line) {

    return (BusConnectionVO) getSession()
        .createQuery(
            "SELECT c FROM BusConnectionVO AS c WHERE" + " c.departureTime = :departure AND c.arrivalTime = :arrival AND"
                + " c.timetable.departureStation.id = :departureStation AND"
                + " c.timetable.arrivalStation.id = :arrivalStation AND"
                + " c.timetable.line.id = :line AND c.supportFor is null").setParameter("departure", departure)
        .setParameter("arrival", arrival).setParameter("departureStation", departureStation.getId())
        .setParameter("arrivalStation", arrivalStation.getId()).setParameter("line", line.getId()).uniqueResult();
  }

  
  @Override
  public List<BusConnectionVO> findConnectionByTimeRange(TimeRange range, LineGroupVO lg) {

    List<WeekDay> days = Arrays.asList(WeekDay.values());
    int day = range.getLower().getDayOfWeek();
    Integer week = DateFormatter.getWeekMod2(range.getLower());
    if(!lg.isInterstate())
      week = DateFormatter.getWeekMod4(range.getLower());
    //zjistim jestli existuje pro dany den nejaka vyjimka (count vrati nenulovou hodnotu). Pokud existuje jakakoli vyjimka na dany den pak hledam jen mezi vyjimkama.
    Long count = (Long) getSession()
        .createQuery(
            "SELECT count(t) FROM BusTurnsTimetableVO as t" + " where t.weekDay = :day"
                + " AND (t.validity_from <= :validity AND t.validity_to >= :validity)").setParameter("day", days.get(day - 1))
        .setParameter("validity", range.getLower()).uniqueResult();
    if (count != 0) {
      return getSession()
          .createQuery(
              "SELECT c FROM BusConnectionVO AS c" + " left join c.ways as w"
                  + " join c.timetable as timetable" + " left join timetable.turns as turns" + " WHERE"
                  + " c.departureTime >= :from AND c.departureTime <= :to" + " AND c.timetable.line.id = :line"
                  + " AND turns.weekDay = :day" + " AND (turns.validity_from <= :validity AND turns.validity_to >= :validity)"
                  + " AND c.supportFor is null" + " order by turns.coupleNumber, c.departureTime")
          .setParameter("from", range.getLower()).setParameter("to", range.getUpper())
          .setParameter("line", lg.getId()).setParameter("day", days.get(day - 1)).setParameter("validity", range.getLower()).list();
    }
    else {
      return getSession()
          .createQuery(
              "SELECT c FROM BusConnectionVO AS c" + 
                  " left join c.ways as w " + 
                  " join c.timetable as timetable"
                  + " left join timetable.turns as turns" + 
                  " WHERE"
                  + " c.departureTime >= :from AND c.departureTime <= :to" + " AND c.timetable.line.id = :line"
                  + " AND turns.weekDay = :day" + " AND turns.week = :week" + " AND (turns.validity_from is null)"
                  + " AND c.supportFor is null" + " order by turns.coupleNumber, c.departureTime")
          .setParameter("from", range.getLower()).
          setParameter("to", range.getUpper()).
          setParameter("line", lg.getId())
          .setParameter("day", days.get(day - 1)).
          setParameter("week", week).list();
    }
  }

  @Override
  public void save(BusConnectionVO conn) {
    if (conn.getTimetable() != null) {
      getSession().save(conn);
      log.info("Connection " + conn.getId() + " was saved");
    }
    else
      log.error("bus_connection " + conn.getId() + " is without timetable");
  }

  @Override
  public List<BusConnectionVO> findSupportsByTimeRange(DateTime day, Long groupLine) {

    DateTime depLoc = DateFormatter.toBeginOfDay(day);
    DateTime arrLoc = DateFormatter.toEndOfDay(day);

    return getSession()
        .createQuery(
            "SELECT distinct c FROM BusConnectionVO AS c" + 
                " left join c.ways" + " join c.timetable as timetable"+
                " left join timetable.turns as turns" + 
                " WHERE"+
                " c.departureTime >= :departure AND c.arrivalTime <= :arrival" + 
                " AND c.timetable.line.id = :line"+
                " AND c.supportFor is not null")
                .setParameter("departure", depLoc)
                .setParameter("arrival", arrLoc)
                .setParameter("line", groupLine).list();
  }

  @Override
  public BusConnectionVO findFirstEmpty(LineGroupVO lg) {
    return (BusConnectionVO) getSession()
        .createQuery(
            "select c from BusConnectionVO c join c.timetable t" + " where t.line.id = :id" + " AND c.ways IS EMPTY"
                + " AND c.departureTime >= :depTime" + " order by c.departureTime asc").setParameter("id", lg.getId())
        .setParameter("depTime", new DateTime()).setMaxResults(1).uniqueResult();
  }
}
