package cz.sa.shiftPlanner.daoImpl;

import java.util.List;

import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.springframework.stereotype.Repository;

import cz.sa.shiftPlanner.dao.PersonDAO;
import cz.sa.shiftPlanner.entities.CityVO;
import cz.sa.shiftPlanner.entities.PersonVO;
import cz.sa.shiftPlanner.entities.StationVO;
import cz.sa.shiftPlanner.infrastructure.dao.AbstractDao;
import cz.sa.shiftPlanner.utils.DateFormatter;

@Repository
public class PersonDAOImpl extends AbstractDao implements PersonDAO {

  private static final Logger logger = Logger.getLogger(PersonDAOImpl.class);
  
  @Override
  public PersonVO get(Long id) {
    return (PersonVO) getSession().get(PersonVO.class, id);
  }

  @Override
  public void create(PersonVO person) {
    if (person == null)
      throw new IllegalArgumentException("person");
    if (person.getId() != null)
      throw new IllegalArgumentException("Already persisted");

    getSession().save(person);
    logger.info("person " + person.getPersonalNumber() + " was created");
  }

  @Override
  public void update(PersonVO person) {
    if (person == null)
      throw new IllegalArgumentException("person");
    if (person.getId() == null)
      throw new IllegalArgumentException("instance not persisted");
    
//    List<CityVO> c = person.getCity();
//    person.setCity(new ArrayList<CityVO>());
//    getSession().update(person);
//    person.setCity(c);
    getSession().update(person);
  }

  @Override
  public void remove(PersonVO person) {
    if (person == null)
      throw new IllegalArgumentException("person");
    if (person.getId() == null)
      throw new IllegalArgumentException("instance not persisted");
    getSession().delete(person);
  }

  @Override
  public List<PersonVO> findAll() {
    return getSession().createQuery(
        "select distinct p from PersonVO as p join p.user as u where u.role LIKE 'steward' order by p.surname").list();
  }

  @Override
  public List<PersonVO> findByName(String name, String surname) {
    if (name == null && surname == null)
      throw new IllegalArgumentException("name and surname is null");
    if (name == null)
      return getSession().createQuery("SELECT p FROM PersonVO AS p WHERE p.surname =:surname").setParameter("surname", surname)
          .list();
    if (surname == null)
      return getSession().createQuery("SELECT p FROM PersonVO AS p WHERE p.name = :name").setParameter("name", name).list();
    return getSession().createQuery("SELECT p FROM PersonVO AS p WHERE p.name = :name AND p.surname =:surname")
        .setParameter("name", name).setParameter("surname", surname).list();
  }

  @Override
  public PersonVO findByPersonalNumber(Long personalNumber) {
    log.info("finding person by personalNumber: " + personalNumber);
    return (PersonVO) getSession().createQuery("SELECT p FROM PersonVO AS p" + " WHERE p.personalNumber =:number")
        .setParameter("number", personalNumber).uniqueResult();
  }

  @Override
  public List<PersonVO> findByLineGroup(Long id) {
    if (id == null)
      throw new IllegalArgumentException("person");
    return getSession().createQuery("SELECT p from PersonVO AS p join p.line_group as l where l.id = :line ")
        .setParameter("line", id).list();
  }

  @Override
  public PersonVO findByParameters(String name, String surname, String mail, String phone) {

    return (PersonVO) getSession()
        .createQuery(
            "SELECT p FROM PersonVO AS p WHERE" + " p.name = :name" + " AND p.surname =:surname"
                + " AND (p.email = :email OR p.phonenumber = :phone)").setParameter("name", name)
        .setParameter("surname", surname).setParameter("email", mail).setParameter("phone", phone).uniqueResult();
  }

  @Override
  public List<PersonVO> findByStations(List<StationVO> stations) {

    return getSession()
        .createQuery(
            "SELECT distinct p from PersonVO AS p" + " left join p.city as c" + " join c.stations as s" + " join p.user as u"
                + " where u.role LIKE 'steward' AND" + " s in (:stations) order by p.surname")
        .setParameterList("stations", stations).list();
  }

  @Override
  public List<PersonVO> findAllByTrialDay(DateTime day) {

    DateTime from = DateFormatter.toBeginOfDay(day);
    DateTime to = DateFormatter.toEndOfDay(day);

    return getSession()
        .createQuery(
            "select distinct p from PersonVO as p join p.timeTrials as t" + " WHERE (t.from_ <= :from AND t.to_ >= :to) OR" + //interval casovky z obou stran presahuje interval from-to
                " (t.from_ >= :from AND t.to_ <= :to) OR" + //interval casovky je cely vnoreny v intervalu from-to
                " (t.from_ <= :from AND t.to_ <= :to) OR" + //interval casovky presahuje jen spodni hranici intervalu from-to
                " (t.from_ >= :from AND t.to_ >= :to)") //interval casovky presahuje jen horni hranici intervalu from-to
        .setParameter("from", from).setParameter("to", to).list();
  }

  @Override
  public List<PersonVO> findByDepCity(CityVO city) {
    return getSession()
        .createQuery(
            "SELECT distinct p from PersonVO AS p" + " left join p.city as c" + " join p.user as u"
                + " where u.role LIKE 'steward' AND" + " c = :city order by p.surname").setEntity("city", city).list();
  }

  @Override
  public List<PersonVO> findByHPP(String jobTime) {
    boolean hpp = false;
    if (jobTime.equalsIgnoreCase("hpp"))
      hpp = true;
    return getSession()
        .createQuery(
            "SELECT distinct p from PersonVO AS p" + " join p.user as u" + " where u.role LIKE 'steward' AND"
                + " p.hpp = :hpp order by p.surname").setBoolean("hpp", hpp).list();
  }

}
