package cz.sa.shiftPlanner.servicesImpl;

import java.util.ArrayList;
import java.util.List;

import org.dozer.Mapper;
import org.joda.time.DateTime;
import org.joda.time.Hours;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import cz.sa.shiftPlanner.dao.PersonUtilsDAO;
import cz.sa.shiftPlanner.dao.WayDAO;
import cz.sa.shiftPlanner.dto.ExcuseDTO;
import cz.sa.shiftPlanner.dto.PersonDTO;
import cz.sa.shiftPlanner.dto.PersonRecordDTO;
import cz.sa.shiftPlanner.dto.TimeTrialDTO;
import cz.sa.shiftPlanner.dto.WayDTO;
import cz.sa.shiftPlanner.entities.ExcuseVO;
import cz.sa.shiftPlanner.entities.PersonRecordVO;
import cz.sa.shiftPlanner.entities.PersonVO;
import cz.sa.shiftPlanner.entities.RuleVO;
import cz.sa.shiftPlanner.entities.TimeTrialVO;
import cz.sa.shiftPlanner.entities.WayVO;
import cz.sa.shiftPlanner.entities.WorkHoursVO;
import cz.sa.shiftPlanner.enums.ExcuseReason;
import cz.sa.shiftPlanner.enums.ExcuseStatus;
import cz.sa.shiftPlanner.services.PersonUtilsService;
import cz.sa.shiftPlanner.services.WayService;
import cz.sa.shiftPlanner.utils.TimeRange;

/**
 * @author ToSk
 */

@Service
@Transactional(propagation = Propagation.REQUIRED)
public class PersonUtilsServiceImpl implements PersonUtilsService {

	private Mapper mapper;
	private PersonUtilsDAO personUtilsDAO;
	private WayService wayService;
	

	public void setWayService(WayService wayService) {
		this.wayService = wayService;
	}

	public void setMapper(Mapper mapper) {
		this.mapper = mapper;
	}

	public void setPersonUtilsDAO(PersonUtilsDAO personUtilsDAO) {
		this.personUtilsDAO = personUtilsDAO;
	}

	@Override
	public void createExcuse(ExcuseDTO excuse) {
		if(excuse == null) throw new IllegalArgumentException("excuse is null");
		ExcuseVO r = new ExcuseVO();
		r = mapper.map(excuse, ExcuseVO.class);
		personUtilsDAO.createExcuse(r);
		mapper.map(r, excuse);
		
		List<WayDTO> list = wayService.findByTimeRange(excuse.getFrom_(), excuse.getTo_(), excuse.getPerson());
		for(WayDTO w : list){
			wayService.archive(w, excuse);
		}
	}
	
  @Override
  public void askExcuse(ExcuseDTO excuse) {
    if(excuse == null) throw new IllegalArgumentException("excuse is null");
    ExcuseVO r = new ExcuseVO();
    r = mapper.map(excuse, ExcuseVO.class);
    personUtilsDAO.createExcuse(r);
    mapper.map(r, excuse);
  }
	
	@Override
	public void remove(ExcuseDTO ex) {
	  personUtilsDAO.remove(mapper.map(ex, ExcuseVO.class));
	}
	
	@Override
	public void updateExcuse(ExcuseDTO ex, ExcuseStatus status) {
	  ex.setStatus(status);
	  personUtilsDAO.update(mapper.map(ex, ExcuseVO.class));
	  
	  if(!status.equals(ExcuseStatus.REJECTED)){
  	  List<WayDTO> list = wayService.findByTimeRange(ex.getFrom_(), ex.getTo_(), ex.getPerson());
      for(WayDTO w : list){
        wayService.archive(w, ex);
      }
	  }
	}

	@Override
	@Transactional(readOnly = true)
	public List<ExcuseDTO> findAll() {
		List<ExcuseDTO> list = new ArrayList<ExcuseDTO>();
		for(ExcuseVO e : personUtilsDAO.findAll()){
			list.add(mapper.map(e, ExcuseDTO.class));
		}
		return list;
	}

	@Override
	@Transactional(readOnly = true)
	public List<ExcuseDTO> findAllByPersonId(Long id) {
		List<ExcuseDTO> list = new ArrayList<ExcuseDTO>();
		for(ExcuseVO e : personUtilsDAO.findAllByPersonId(id)){
			list.add(mapper.map(e, ExcuseDTO.class));
		}
		return list;
	}
	
	@Override
	@Transactional(readOnly = true)
  public List<ExcuseDTO> excusesInDay(DateTime dt, ExcuseReason reason) {
	  List<ExcuseDTO> list = new ArrayList<ExcuseDTO>();
    for(ExcuseVO e : personUtilsDAO.excusesInDay(dt, reason)){
      list.add(mapper.map(e, ExcuseDTO.class));
    }
    return list;
	}

	/**
	 * calculate month hours for set mont and year
	 * @param person is for who method count
	 * @param month 
	 * @param year
	 * @return sum in hours
	 */
	@Override
	public Integer sumMonthHours(PersonDTO person, int month, int year) {
		
		WorkHoursVO vo = personUtilsDAO.findHourByIdAndMonth(person.getId(), month, year);
		if(vo == null) return 0;
		return vo.getMonthAmount();
	}

	@Override
	public Long limitMonthHours(PersonDTO person, int month, int year) {

		WorkHoursVO vo = personUtilsDAO.findHourByIdAndMonth(person.getId(), month, year);
		if(vo == null) return 0L;
		return Long.valueOf(vo.getHoursNorm());
	}


	@Override
	public List<TimeTrialDTO> findAllByPersonId(long personId, int month, int year) {
		
		DateTime from = new DateTime(year, month, 1, 0, 0, 0, 0);
		//pridam jeden mesic, ale odectu den abych docilil posledniho dne v mesici
		DateTime to = from.plusMonths(1);
		to = to.minusDays(1);
		List<TimeTrialDTO> list = new ArrayList<TimeTrialDTO>();
		for(TimeTrialVO e : personUtilsDAO.findAllByPersonId(personId , from, to)){
			list.add(mapper.map(e, TimeTrialDTO.class));
		}
		return list;
	}
	
	@Override
	public void createTrials(List<TimeRange> ranges, PersonDTO person) {
	  PersonVO p = mapper.map(person, PersonVO.class);
	  for (TimeRange tr : ranges) {
	    TimeTrialVO trial = new TimeTrialVO();
	    trial.setFrom_(tr.getLower());
	    trial.setTo_(tr.getUpper());
	    trial.setPerson(p);
	    trial.setNote(null);
	    personUtilsDAO.create(trial);
    }
	}

	@Override
	public void createTrial(TimeTrialDTO tt) {
		personUtilsDAO.create(mapper.map(tt, TimeTrialVO.class));
	}

	@Override
	public void updateTrial(TimeTrialDTO tt) {
		personUtilsDAO.update(mapper.map(tt, TimeTrialVO.class));
	}

	@Override
	public void remove(TimeTrialDTO tt) {
		personUtilsDAO.remove(mapper.map(tt, TimeTrialVO.class));
	}

	@Override
	public List<TimeTrialDTO> findAllByDay(DateTime day) {
		
		List<TimeTrialDTO> list = new ArrayList<TimeTrialDTO>();
		for(TimeTrialVO e : personUtilsDAO.findAllByDay(day)){
			list.add(mapper.map(e, TimeTrialDTO.class));
		}
		return list;
	}

  @Override
  public void createRecord(PersonRecordDTO record) {
    personUtilsDAO.createRecord(mapper.map(record, PersonRecordVO.class));
  }

  @Override
  public List<PersonRecordDTO> findRecordsByPerson(PersonDTO person) {
    List<PersonRecordDTO> list = new ArrayList<PersonRecordDTO>();
    for(PersonRecordVO e : personUtilsDAO.findRecordByPerson(mapper.map(person, PersonVO.class))){
      list.add(mapper.map(e, PersonRecordDTO.class));
    }
    return list;
  }

  @Override
  public void removeRecord(PersonRecordDTO rec) {
    personUtilsDAO.removeRecord(mapper.map(rec, PersonRecordVO.class));
  }

}


