package cz.sa.shiftPlanner.servicesImpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.dozer.Mapper;
import org.joda.time.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import cz.sa.shiftPlanner.dao.PersonUtilsDAO;
import cz.sa.shiftPlanner.dao.WayDAO;
import cz.sa.shiftPlanner.dto.BusConnectionDTO;
import cz.sa.shiftPlanner.dto.BusTurnsTimetableDTO;
import cz.sa.shiftPlanner.dto.ExcuseDTO;
import cz.sa.shiftPlanner.dto.NotificationDTO;
import cz.sa.shiftPlanner.dto.PersonDTO;
import cz.sa.shiftPlanner.dto.WayDTO;
import cz.sa.shiftPlanner.entities.ExcuseVO;
import cz.sa.shiftPlanner.entities.PersonVO;
import cz.sa.shiftPlanner.entities.WayVO;
import cz.sa.shiftPlanner.entities.WayVOAudit;
import cz.sa.shiftPlanner.entities.WorkHoursItemVO;
import cz.sa.shiftPlanner.entities.WorkHoursVO;
import cz.sa.shiftPlanner.enums.NotifType;
import cz.sa.shiftPlanner.enums.WeekDay;
import cz.sa.shiftPlanner.services.BusTurnsTimetableService;
import cz.sa.shiftPlanner.services.NotificationService;
import cz.sa.shiftPlanner.services.WayService;
import cz.sa.shiftPlanner.utils.DateFormatter;
import cz.sa.shiftPlanner.utils.ServiceNoticeException;
import cz.sa.shiftPlanner.utils.TimeRange;
import cz.sa.shiftPlanner.utils.WayDTOComparator;
import cz.sa.shiftPlanner.utils.WayDTOListComparator;

@Service
@Transactional(propagation = Propagation.REQUIRED)
public class WayServiceImpl implements WayService {

	private WayDAO wayDAO;
	private Mapper mapper;
	private PersonUtilsDAO personUtilsDAO;
	private BusTurnsTimetableService busTurnsTimetableService;
	
	@Autowired
	@Qualifier("notificationService")
	NotificationService notificationService;
	
	private final int CLEARANCE = 15;
	private final int BREAK_TIME = 30;
	
	public void setBusTurnsTimetableService(
			BusTurnsTimetableService busTurnsTimetableService) {
		this.busTurnsTimetableService = busTurnsTimetableService;
	}

	public void setPersonUtilsDAO(PersonUtilsDAO personUtilsDAO) {
		this.personUtilsDAO = personUtilsDAO;
	}

	public void setMapper(Mapper mapper) {
		this.mapper = mapper;
	}
	
	public void setWayDAO(WayDAO wayDAO) {
		this.wayDAO = wayDAO;
	}

	@Override
	public WayDTO get(Long id) {
		return mapper.map(wayDAO.get(id),WayDTO.class);
	}

	@Override
	public void create(WayDTO way) throws ServiceNoticeException {
		if(way == null) throw new IllegalArgumentException("way on service is null");
		
		checkExcuse(way.getConnection(), way.getPerson());
		checkTimeConflict(way.getConnection(), way.getPerson());
		//zkontroluje prekroceni hodinoveho limitu a pripadne zapise dalsi hodiny
		checkPersonWorkHours(new TimeRange(way.getConnection().getDepartureTime(), way.getConnection().getArrivalTime()), way.getPerson(), false, true);
		
		WayVO w = mapper.map(way,WayVO.class);
		wayDAO.create(w);
		mapper.map(w, way);
	}
	
	@Override
  public void createUseWayTime(WayDTO way) throws ServiceNoticeException {
    if(way == null) throw new IllegalArgumentException("way on service is null");
    
    checkExcuse(way.getConnection(), way.getPerson());
    checkTimeConflict(way.getConnection(), way.getPerson());
    //zkontroluje prekroceni hodinoveho limitu a pripadne zapise dalsi hodiny
    checkPersonWorkHours(new TimeRange(way.getDepartureTime(), way.getArrivalTime()), way.getPerson(), false, true);
    
    WayVO w = mapper.map(way,WayVO.class);
    wayDAO.create(w);
    mapper.map(w, way);
  }
	
	@Override
	public void update(WayDTO way) {
		wayDAO.update(mapper.map(way,WayVO.class));
	}
	
	@Override
  public void archive(WayDTO way, ExcuseDTO excuse) {
	  remove(way);
    
    WayVOAudit audit = mapper.map(way, WayVOAudit.class);
    
    audit.setExcuse(mapper.map(excuse, ExcuseVO.class));
    audit.setId(null);
    wayDAO.createAuditVO(audit);
  }
	

	@Override
	public void remove(WayDTO way) {
		WorkHoursVO w = personUtilsDAO.findHourByIdAndMonth(way.getPerson().getId(), way.getDepartureTime().getMonthOfYear(), way.getDepartureTime().getYear());
		if(w==null) return;
		WorkHoursItemVO it = personUtilsDAO.findItemByIdAndDay(w.getId(), way.getDepartureTime().getDayOfMonth());
		if(it==null) return;
		int newAmount = it.getDay_amount() - (parseTimesToInterval(way.getDepartureTime(), way.getArrivalTime())+BREAK_TIME);
		if(newAmount < 0) throw new IllegalArgumentException("amount of minutes in "+way.getDepartureTime()+" is less zero");
		it.setDay_amount(newAmount);
		personUtilsDAO.updateWorkHoursItem(it);
		wayDAO.remove(wayDAO.get(way.getId()));
	}
	
	public void remove(WayDTO way, boolean clearance) {
		WorkHoursVO w = personUtilsDAO.findHourByIdAndMonth(way.getPerson().getId(), way.getDepartureTime().getMonthOfYear(), way.getDepartureTime().getYear());
		if(w==null) return;
		WorkHoursItemVO it = personUtilsDAO.findItemByIdAndDay(w.getId(), way.getDepartureTime().getDayOfMonth());
		if(it==null) return;
		int newAmount = 0;
		//pokud se jedna o mazani odbaveni ubira se jen 15 minut
		if(clearance) 
			newAmount = it.getDay_amount() - CLEARANCE;
		else 
			newAmount = it.getDay_amount() - (parseTimesToInterval(way.getDepartureTime(), way.getArrivalTime())+BREAK_TIME);
			//if(newAmount < 0) throw new IllegalArgumentException("amount of minutes in "+way.getDepartureTime()+" is less zero");
		it.setDay_amount(newAmount);
		personUtilsDAO.updateWorkHoursItem(it);
		wayDAO.remove(wayDAO.get(way.getId()));
	}
	
	@Override
	public WayDTO create(BusConnectionDTO conn, PersonDTO per, boolean clearance, boolean isAdmin) throws ServiceNoticeException {

		checkExcuse(conn, per);
		//pokud se jedna o odbaveni neprovadi se kontrola na casovy konflikt, protoze by se kontrolovala cela cesta autobusu i kdyz s nim stew nejede
		if(!clearance)
			checkTimeConflict(conn, per);
		//zkontroluje jestli nechochazi k preteceni hodin. Pokud zapisuje admin a jsou prekroceny hodiny, pak mu to dovoli a vyjimku nevychazuje
		if(!checkPersonWorkHours(new TimeRange(conn.getDepartureTime(), conn.getArrivalTime()), per, clearance, isAdmin) && !isAdmin) 
			throw new ServiceNoticeException("Prekrocili byste mesicni hodinovy limit, spoj nebyl pridan do databaze!");
		WayDTO way = new WayDTO();
		way.setArrivalStation(conn.getTimetable().getArrivalStation());
		way.setArrivalTime(conn.getArrivalTime());
		way.setDepartureStation(conn.getTimetable().getDepartureStation());
		way.setDepartureTime(conn.getDepartureTime());
		way.setDelay(0);
		way.setConnection(conn);
		way.setPerson(per);
		way.setClearance(clearance);
		if(way.getArrivalStation().getCity().getCountry().equalsIgnoreCase("cz"))
			way.setIsForeign(false);
		else
			way.setIsForeign(true);
		WayVO w = mapper.map(way,WayVO.class);
		wayDAO.create(w);
		mapper.map(w, way);
		
		if(!isAdmin){
			NotificationDTO note = new NotificationDTO();
			note.setCreated(new DateTime());
			note.setReaded(false);
			note.setPerson(per);
			note.setNotifType(NotifType.STEWARD);
			note.setText("Vytvorena nova cesta z "+ way.getDepartureStation().getCity()+" "+ DateFormatter.toDateAndTime(way.getDepartureTime())+" do " + way.getArrivalStation().getCity()+" "+ DateFormatter.toDateAndTime(way.getArrivalTime()));
			notificationService.create(note);
		}
		return way;
	}

	private void checkExcuse(BusConnectionDTO conn, PersonDTO per) throws ServiceNoticeException {

		List<TimeRange> list = personUtilsDAO.rangesInDay(conn.getDepartureTime(), mapper.map(per, PersonVO.class));
		
		for(TimeRange r : list){
			if(r.overlap(conn.getDepartureTime(), conn.getArrivalTime()))
				throw new ServiceNoticeException("Spoj(e) nelze zapsat z duvodu nahlasene absence.");
		}
	}

	private void checkTimeConflict(BusConnectionDTO conn, PersonDTO per) throws ServiceNoticeException {
		
		List<WayVO> list = wayDAO.findByDepRange(conn.getDepartureTime(), conn.getArrivalTime(), mapper.map(per, PersonVO.class));
		if(list.size() != 0) 
			throw new ServiceNoticeException(
					"Na lince \""+list.get(0).getConnection().getTimetable().getLine().getName()+
					"\" je v rozmezi "+
					DateFormatter.toDateAndTime(conn.getDepartureTime())+
					" - "+
					DateFormatter.toDateAndTime(conn.getArrivalTime())+
					" konflikt s jinym spojem.");
	}

	/**
	 * 
	 * @param range where lower means departureTime and upper means ArrivalTime
	 * @param per
	 * @param clearance
	 * @param isAdmin
	 * @return
	 */
	private boolean checkPersonWorkHours(TimeRange range, PersonDTO per, boolean clearance, boolean isAdmin) {
		//pokusim se najit hodinovy zaznam pro mesic a rok
		WorkHoursVO hours = personUtilsDAO.findHourByIdAndMonth(per.getId(), range.getUpper().getMonthOfYear(), range.getUpper().getYear());
		WorkHoursItemVO item;
		int minutesAmount = parseTimesToInterval(range.getLower(), range.getUpper()) + BREAK_TIME;
		//pokud se jedna jen o odbaveni zapocitava se jen 15 minut
		if(clearance) minutesAmount= CLEARANCE;
		//pokud nenajdu, pak vytvorim novy zaznam vcetne zaznamu pro dany den
		if(hours == null){
			hours = new WorkHoursVO();
			hours.setMonth(range.getLower().getMonthOfYear());
			hours.setPerson(mapper.map(per, PersonVO.class));
			hours.setYear(range.getLower().getYear());
			//zjistim si kolik je dnu v mesici
			int workDays = DateFormatter.getDaysInMonth(range.getLower());
			hours.setHoursNorm(((per.getWeekWorkTime()*60)/7)*workDays);
			personUtilsDAO.createWorkHours(hours);
			
			item = new WorkHoursItemVO();
			item.setDay_amount(minutesAmount);
			item.setDayOfMonth(range.getLower().getDayOfMonth());
			item.setWorkHours(hours);
			personUtilsDAO.createWorkHoursItem(item);
			return true;
		//pokud jsem zaznam masel, pak se podivam jestli existuje v mesici zaznam pro aktualni den. 
		//zkontroluju jestli prirustek hodin odpovida norme a pak pripadne vytvorim nebo updatuju zaznam pro urcity den.
		}else{
			item = personUtilsDAO.findItemByIdAndDay(hours.getId(), range.getLower().getDayOfMonth());
			if(item == null){
				if(checkPersonMonthWorkHours(minutesAmount, hours) || isAdmin){
					item = new WorkHoursItemVO();
					item.setDay_amount(minutesAmount);
					item.setDayOfMonth(range.getLower().getDayOfMonth());
					item.setWorkHours(hours);
					personUtilsDAO.createWorkHoursItem(item);
					return true;
				}
			}else{
				if(checkPersonMonthWorkHours(minutesAmount, hours) || isAdmin){
					int newAmount = item.getDay_amount() + minutesAmount;
					item.setDay_amount(newAmount);
					personUtilsDAO.updateWorkHoursItem(item);
					return true;
				}
			}
		}
		return false;
	}
	
	
	
	private boolean checkPersonMonthWorkHours(int amountPlus, WorkHoursVO hours) {
		
		if((personUtilsDAO.sumMonthHours(hours.getId()) + amountPlus) <= (hours.getHoursNorm()))
			return true;
		else return false;
	}
	
	private int parseTimesToInterval(DateTime dep, DateTime arr){
		
		Long duration = arr.getMillis() - dep.getMillis();
		return (int)java.util.concurrent.TimeUnit.MILLISECONDS.toMinutes(duration);
	}

	

	@Override
	@Transactional(propagation = Propagation.REQUIRED, readOnly=true)
	public List<WayDTO> findByTimeRange(DateTime from, DateTime to, PersonDTO p) {
		List<WayDTO> list = new ArrayList<WayDTO>();
		for(WayVO w : wayDAO.findByTimeRange(from, to, mapper.map(p, PersonVO.class))){
			list.add(mapper.map(w, WayDTO.class));
		}
		return list;
	}

	@Override
	@Transactional(propagation = Propagation.REQUIRED, readOnly=true)
	public List<List<WayDTO>> findByMonth(int month, PersonDTO p) {

		DateTime from = new DateTime(new DateTime().getYear(), month, 1, 0, 0, 0, 0);
		//pridam jeden mesic, ale odectu den abych docilil posledniho dne v mesici
		DateTime to = from.plusMonths(1);
		to = to.minusDays(1);
		List<WayDTO> list = new ArrayList<WayDTO>();
		for(WayVO w : wayDAO.findByMonth(from, to, mapper.map(p, PersonVO.class))){
		  WayDTO dto = mapper.map(w, WayDTO.class);
      dto.setIsActive(true);
      list.add(dto);
		}
		for(WayVOAudit w : wayDAO.findByMonthNonActive(from, to, mapper.map(p, PersonVO.class))){
      WayDTO dto = mapper.map(w, WayDTO.class);
      dto.setIsActive(false);
      list.add(dto);
    }
		if(list.size()==0) return new ArrayList<List<WayDTO>>();
		
		List<List<WayDTO>> result = sortListToLists(list);
		Collections.sort(result, new WayDTOListComparator());
		return result;
	}
	
	private List<List<WayDTO>> sortListToLists(List<WayDTO> list) {

		List<WeekDay> days = Arrays.asList(WeekDay.values());
		
		List<List<WayDTO>> result = new ArrayList<List<WayDTO>>();
		if(list.size()==0) return result;
	
		DateTime depTime = list.get(0).getDepartureTime();
		
		BusTurnsTimetableDTO b = busTurnsTimetableService.findByDayAndTimetable(
				list.get(0).getConnection().getTimetable(),
				days.get(depTime.getDayOfWeek()-1),
				DateFormatter.toBeginOfDay(depTime),
				DateFormatter.getWeekMod2(depTime)
				);
		if(b==null) return result;
		int coupleNumber = b.getCoupleNumber();
			
		List<WayDTO> listLoc = new ArrayList<WayDTO>();
		
		for(WayDTO dto : list){
			// pokud se objevi bez couple number nebo bez Turn. Pak to znamena chybu a je treba tento spoj nahlasit 
			DateTime depTimeLoc = dto.getDepartureTime();
			BusTurnsTimetableDTO turn = busTurnsTimetableService.findByDayAndTimetable(
					dto.getConnection().getTimetable(),
					days.get(depTimeLoc.getDayOfWeek()-1),
					DateFormatter.toBeginOfDay(depTimeLoc),
					DateFormatter.getWeekMod2(depTimeLoc)
					);
			Integer coupleNumberLoc = turn == null ? null : turn.getCoupleNumber();
			if(coupleNumberLoc == null)
				 throw new IllegalArgumentException("problem with couplenumber");
			dto.getConnection().setBusTurn(turn);
			if(coupleNumber == coupleNumberLoc && depTime.getDayOfMonth() == depTimeLoc.getDayOfMonth()){
				if(!listLoc.contains(dto))
					listLoc.add(dto);
			}else{
			  Collections.sort(listLoc, new WayDTOComparator()); 
				result.add(listLoc);
				coupleNumber = coupleNumberLoc;
				depTime= depTimeLoc;
				listLoc = new ArrayList<WayDTO>();
				listLoc.add(dto);
			}
		}
		if(!result.contains(listLoc)){
		  Collections.sort(listLoc, new WayDTOComparator()); 
		  result.add(listLoc);
		}
		return result;
	}

	@Override
	public void createAll(List<BusConnectionDTO> conn, PersonDTO per, boolean clearance, boolean isAdmin) throws ServiceNoticeException {
		for(BusConnectionDTO busT : conn){
			if(busT.getWays().size()==0){
				busT.getWays().add(create(busT, per, clearance, isAdmin));
			}
		}
	}

  @Override
  public List<WayDTO> findAll() {
    List<WayDTO> list = new ArrayList<WayDTO>();
    for(WayVO w : wayDAO.findAll()){
      list.add(mapper.map(w, WayDTO.class));
    }
    return list;
  }

}
